(*----------------------------------------------------------------------------*
 *  Direct3D sample from DirectX 9.0 SDK Summer 2003                          *
 *  Delphi adaptation by Alexey Barkovoy (e-mail: clootie@ixbt.com)           *
 *                                                                            *
 *  Delphi versions 5-7 are supported                                         *
 *                                                                            *
 *  Modified: 10-Jan-2004                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru/delphi                                         *
 *----------------------------------------------------------------------------*)
//-----------------------------------------------------------------------------
// File: BumpSelfShadow.cpp
//
// Desc: Shows how to implement self-shadowing bump mapping
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------

unit BumpSelfshadowUnit;

interface

uses
  Windows, Messages, SysUtils, Math, CommDlg,
  Direct3D9, D3DX9, D3DApp, D3DFont, D3DUtil, DXUtil;

//-----------------------------------------------------------------------------
// Defines, constants, and global variables
//-----------------------------------------------------------------------------
type

  // A complex vertex, if we use vertex shaders we can handle this data ourselves
  TBiNormalVertex = record
    x, y, z:    Single;
    u, v:       Single;
    nx, ny, nz: Single;
    bx, by, bz: Single;
    col:        DWord;
    RefCount:   Integer;
  end;

  TShaderVertex = record
    x, y, z:    Single;
    u, v:       Single;
    nx, ny, nz: Single;
    bx, by, bz: Single;
  end;

  TNormalVertex = record
    x, y, z:    Single;
    u, v:       Single;
    nx, ny, nz: Single;
  end;


  TBasicVertex = record
    x, y, z:    Single;
    col:        DWord;
    u,  v:      Single;
  end;

const
  // BasicVertex2::FVF = D3DFVF_XYZ | D3DFVF_DIFFUSE | D3DFVF_TEXCOORDSIZE2(0) | D3DFVF_TEXCOORDSIZE2(1);
  TBasicVertex_FVF = D3DFVF_XYZ or D3DFVF_DIFFUSE or D3DFVF_TEX1;

type
  //A BasicVertex FVF2
  TBasicVertex2 = record
    x, y, z:    Single;
    col:        DWord;
    u, v:       Single;
    u1, v2:     Single;
  end;

const
  // FVF_BASICVERTEX2 = D3DFVF_XYZ or D3DFVF_DIFFUSE or D3DFVF_TEXCOORDSIZE2(0) or D3DFVF_TEXCOORDSIZE2(1);
  FVF_BASICVERTEX2 = D3DFVF_XYZ or D3DFVF_DIFFUSE or D3DFVF_TEXTUREFORMAT2;

  MAXTRIANGLES  = 4096;
  BASISMAPSIZE  = 32;
  RENDERTARGSIZE = 256;

type
  TsTriangle = record
    ind1, ind2, ind3: Word;
  end;

type
  TShadowSet = class
    m_Triangles: array[0..MAXTRIANGLES-1] of TsTriangle;
    m_pScreenVB: IDirect3DVertexBuffer9;
    m_pShad1:    IDirect3DVertexBuffer9;
    m_pVertices: IDirect3DVertexBuffer9;

    m_TriVertices: array[0..MAXTRIANGLES*3 - 1] of TBiNormalVertex;

    m_iNumTriangles, m_iNumVertices: Integer;
    m_iHeightWidth, m_iHeightHeight: DWORD;

    m_pDev: IDirect3DDevice9;

    // our interleaved horizon maps with the precomputed basis textures
    m_pHorizonTextures: array[0..3] of IDirect3DTexture9;
    m_pBasisTextures:   array[0..3] of IDirect3DTexture9;

    // we render into this buffer and use it as a texture
    m_pRenderBuffer:  IDirect3DTexture9;
    m_pRenderBuffer2: IDirect3DTexture9;
    m_pTempBuffer:    IDirect3DTexture9;

    m_pOldRenderTarget: IDirect3DSurface9;

    m_pNormalMap: IDirect3DTexture9;

    m_pColorTexture: IDirect3DTexture9;
    m_pIndexBuffer: IDirect3DIndexBuffer9;

    m_matObject: TD3DXmatrix;
    m_matView:   TD3DXmatrix;
    m_matProj:   TD3DXmatrix;
    m_matTot:    TD3DXmatrix;

    // our shaders
    m_pVertexDeclaration: IDirect3DVertexDeclaration9;
    m_pBasisShader:  IDirect3DVertexShader9;
    m_pShadowShader: IDirect3DVertexShader9;
    m_pBumpShader:   IDirect3DVertexShader9;
    m_pPixShader:    IDirect3DPixelShader9;

    m_cColorName: array[0..159] of Char;
    m_cHeightName: array[0..159] of Char;
    m_dwHeightLevels: DWord;

    m_bInitalize: BOOL;
    m_bLoaded: BOOL;
    m_bUsePix: BOOL;

    // scratch space for loading/computing the textures
    m_pHeightMap:   array[0..15] of PByteArray;
    m_pHorizonMap:  array[0..15] of PByteArray;
    m_pBasisMap:    array[0..15] of PByteArray;
    m_pHorizonComp: array[0..3] of PByte;
    m_pBasisComp:   array[0..3] of PByte;

    // linear basis map
    procedure BuildBasisMap(pBasis: PByteArray; PrimAngle, fAngle2: Single);

    function BuildInterleavedMap(out newTex: IDirect3DTexture9; r, g, b, a: PByteArray;
      width, height: Integer): Boolean;

    procedure BuildHorizonMap(pHor, pHeight: PByteArray; dx, dy: Single;
      iStepLength: Integer; iWidth, iHeight: Integer; bWrapU, bWrapV: Boolean);

  public
    constructor Create(HeightMapName, ColorTexName: PChar);
    destructor Destroy; override;

    procedure ComputeLight(const lightDir: TD3DXVector3; const pObject, pView, pProj: TD3DXMatrix);
    procedure Render(bShow, bShad, bBump: BOOL; iLeft, iTop, iRight, iBottom: Integer);
    procedure SetBuffers(pIndex: IDirect3DIndexBuffer9; pVertBuffer: IDirect3DVertexBuffer9;
      iNumFaces, iNumVertices: Integer);

    function Init(pDev: IDirect3DDevice9; bUsePix: BOOL): HResult;
    procedure DeleteDeviceObj;
    procedure Restore;
    procedure Invalidate;
  end;


  //-----------------------------------------------------------------------------
  // Name: class CMyD3DApplication
  // Desc: Application class. The base class (CD3DApplication) provides the
  //       generic functionality needed in all Direct3D samples. CMyD3DApplication
  //       adds functionality specific to this sample program.
  //-----------------------------------------------------------------------------
  CMyD3DApplication = class(CD3DApplication)
  private
    m_pShadow: TShadowSet;

    m_matObject:        TD3DXMatrix;
    m_matView:          TD3DXMatrix;
    m_matProj:          TD3DXMatrix;
    m_matPosition:      TD3DXMatrix;

    m_LightDir:         TD3DXVector3;
    m_bLMouseDown:      BOOL;
    m_bRMouseDown:      BOOL;
    m_bFirstTime:       BOOL;
    m_bRotate:          BOOL;
    m_xPos, m_yPos:     Single;
    m_oPosX, m_oPosY:   Single;
    m_iIndex:           Integer;

    m_bKey: array[0..255] of Boolean;
    m_fSpeed:           Single;
    m_fAngularSpeed:    Single;
    m_fTheta:           Single;
    m_fTheta2:          Single;

    m_bDebug:           BOOL;
    m_bShadow:          BOOL;
    m_bBump:            BOOL;
    m_bUsePix:          BOOL;

    m_vecVelocity:      TD3DXVector3;
    m_vecAngularVelocity: TD3DXVector3;

  protected
    function OneTimeSceneInit: HResult; override;
    function InitDeviceObjects: HResult; override;
    function RestoreDeviceObjects: HResult; override;
    function InvalidateDeviceObjects: HResult; override;
    function DeleteDeviceObjects: HResult; override;
    function FrameMove: HResult; override;
    function Render: HResult; overload; override;
    function FinalCleanup: HResult; override;
  public
    constructor Create;
    
    function ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
      adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT; override;
    function MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT; override;
  private
    m_fAspectRatio: Single;

    // procedure CreateSphere; //???
    procedure OnMouseMove;
    procedure LoadXFile(name: PChar);

  private    
    m_pFont:            CD3DFont;
    m_pFontSmall:       CD3DFont;
  end;

implementation

{ CMyD3DApplication }

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
var
  i: Integer;
  matT: TD3DXMatrix;
begin
  inherited Create;


  m_fElapsedTime := 0.0;
  m_fAspectRatio := 0.0;

  m_dwCreationWidth     := 400;   // Width used to create window
  m_dwCreationHeight    := 400;

  m_d3dEnumeration.AppUsesDepthBuffer := True;
  m_bShowCursorWhenFullscreen := True;

  m_strWindowTitle      := 'Self-Shadow Bumpmap';

  m_pShadow             := nil;
  m_pFont               := CD3DFont.Create('Arial', 12, D3DFONT_BOLD);
  m_pFontSmall          := CD3DFont.Create('Arial',  9, D3DFONT_BOLD);

  for i:= 0 to 255 do m_bKey[i]:= False;

  m_bLMouseDown := False;
  m_bRMouseDown := False;
  m_bFirstTime  := True;

  m_iIndex := 0;

  m_bDebug      := True;
  m_bShadow     := True;
  m_bBump       := True;

  m_fTheta      := 0;
  m_fTheta2     := 0;

  D3DXMatrixIdentity(m_matObject);
  D3DXMatrixIdentity(m_matView);
  D3DXMatrixIdentity(m_matPosition);

  //initalize the camera and keyboard speeds
  //this should probally change depending on the size
  //of the object being viewed
  m_fSpeed             := 20.00;
  m_fAngularSpeed      :=  0.10;
  m_vecVelocity        := D3DXVector3(0.0, 0.0, 0.0);
  m_vecAngularVelocity := D3DXVector3(0.0, 0.0, 0.0);
  m_LightDir           := D3DXVector3(0.0, 0.0, 1.0);

  D3DXMatrixTranslation(matT, 0,0, -40);

  D3DXMatrixMultiply(m_matPosition, matT, m_matPosition);
  D3DXMatrixInverse(m_matView, nil, m_matPosition);
end;

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device initialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function CMyD3DApplication.ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
  adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT;
begin
  Result:= E_FAIL;

  if ((dwBehavior and D3DCREATE_HARDWARE_VERTEXPROCESSING) <> 0) or
     ((dwBehavior and D3DCREATE_MIXED_VERTEXPROCESSING) <> 0) then
  begin
    if (pCaps.VertexShaderVersion < D3DVS_VERSION(1,0)) then Exit;
  end;

  if ((pCaps.TextureOpCaps and D3DTEXOPCAPS_DOTPRODUCT3) = 0) then Exit;
  if (pCaps.MaxTextureBlendStages < 2) then Exit;

  if (pCaps.PixelShaderVersion < D3DPS_VERSION(1,1)) then
  begin
    if FAILED(m_pD3D.CheckDeviceFormat(pCaps.AdapterOrdinal,
         pCaps.DeviceType, adapterFormat, D3DUSAGE_RENDERTARGET,
         D3DRTYPE_TEXTURE, D3DFMT_A8R8G8B8)) then Exit;
  end;

  // Need to support post-pixel processing (for alpha blending)
  if FAILED(m_pD3D.CheckDeviceFormat(pCaps.AdapterOrdinal, pCaps.DeviceType,
       adapterFormat, D3DUSAGE_RENDERTARGET or D3DUSAGE_QUERY_POSTPIXELSHADER_BLENDING,
       D3DRTYPE_SURFACE, backBufferFormat)) then Exit;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup: HResult;
begin
  SAFE_DELETE(m_pShadow);
  SAFE_DELETE(m_pFont);
  SAFE_DELETE(m_pFontSmall);

  Result := S_OK;
end;

//track mouse movement
//just a crude interface to let the user move the object
procedure CMyD3DApplication.OnMouseMove;
var
  r: TRect;
  dx, dy: Single;
  mat1, mat2, mat3: TD3DXMatrix;
begin
  GetWindowRect(m_hWnd, r);
  dx:= m_xPos / (r.right-r.left)*2 - 1;
  dy:= m_yPos / (r.right-r.left)*2 - 1;

  //right mouse button moves the light around
  if m_bRMouseDown then
  begin
    m_LightDir.x := -dx;
    m_LightDir.y :=  dy;
    m_LightDir.z := 0.25;
  end;

  //left button rotates the object
  if m_bLMouseDown then
  begin
    m_fTheta  := -dy + m_oPosY;
    m_fTheta2 := -dx + m_oPosX;

    D3DXMatrixRotationX(mat1, m_fTheta*0.1);
    D3DXMatrixRotationY(mat2, m_fTheta2*0.1);
    D3DXMatrixMultiply(mat3, mat1, mat2);
    D3DXMatrixMultiply(m_matObject, m_matObject, mat3);
  end;
end;

//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Message proc function to handle key and menu input
//-----------------------------------------------------------------------------
function CMyD3DApplication.MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT;
var
  r: TRect;
//  matT: TD3DXMatrix;
begin
  // Trap the context menu
  if (WM_CONTEXTMENU = uMsg) then
  begin
    Result:= 0;
    Exit;
  end;

  case uMsg of
    WM_KEYUP:
      //mark a key up
      m_bKey[wParam]:= FALSE;

    WM_KEYDOWN:
    begin
      //mark a key down
      m_bKey[wParam]:= TRUE;

       //turn shadowing off
      if (wParam = Ord('2')) then
        m_bShadow := not m_bShadow;

      //turn bump mapping off
      if (wParam = Ord('3')) then
        m_bBump := not m_bBump;

      //auto rotate
      if (wParam = Ord('5')) then
        m_bRotate := not m_bRotate;

      //reset matrices
      if (wParam = Ord('4')) then
      begin
        D3DXMatrixTranslation(m_matPosition, 0,0, -40);
        D3DXMatrixInverse(m_matView, nil, m_matPosition);
        D3DXMatrixIdentity(m_matObject);
      end;
    end;

    //track left mouse button down
    WM_LBUTTONDOWN:
    begin
      SetCapture(m_hWnd);
      m_bLMouseDown := True;
      m_xPos := LOWORD(lParam);
      m_yPos := HIWORD(lParam);

      GetWindowRect(m_hWnd, r);

      m_oPosX := m_xPos / (r.right-r.left)*2 - 1;
      m_oPosY := m_yPos / (r.right-r.left)*2 - 1;
    end;

    //track right mouse button
    WM_RBUTTONDOWN:
    begin
      m_xPos := LOWORD(lParam);
      m_yPos := HIWORD(lParam);

      m_bRMouseDown := True;
    end;

    //track mouse movement
    WM_MOUSEMOVE:
    begin
      if (m_bRMouseDown or m_bLMouseDown) then
      begin
        m_xPos := LOWORD(lParam);
        m_yPos := HIWORD(lParam);
      end;

      OnMouseMove;
    end;

    WM_RBUTTONUP:
      m_bRMouseDown:= False;

    WM_LBUTTONUP:
    begin
      ReleaseCapture;
      m_bLMouseDown:= False;
    end;
  end;

  Result:= inherited MsgProc(hwnd, uMsg, wParam, lParam);
end;

//load an X file
var
  declVS: array[0..4] of TD3DVertexElement9 =
  (
    (Stream: 0; Offset:  0; _Type: D3DDECLTYPE_FLOAT3; Method: D3DDECLMETHOD_DEFAULT; Usage: D3DDECLUSAGE_POSITION; UsageIndex: 0),
    (Stream: 0; Offset: 12; _Type: D3DDECLTYPE_FLOAT3; Method: D3DDECLMETHOD_DEFAULT; Usage: D3DDECLUSAGE_NORMAL; UsageIndex: 0),
    (Stream: 0; Offset: 24; _Type: D3DDECLTYPE_FLOAT2; Method: D3DDECLMETHOD_DEFAULT; Usage: D3DDECLUSAGE_TEXCOORD; UsageIndex: 0),
    (Stream: 0; Offset: 32; _Type: D3DDECLTYPE_FLOAT3; Method: D3DDECLMETHOD_DEFAULT; Usage: D3DDECLUSAGE_TANGENT; UsageIndex: 0),
    ({D3DDECL_END})
  );
procedure CMyD3DApplication.LoadXFile(name: PChar);
var
  pMeshSysMem, pMeshSysMem2: ID3DXMesh;
  pIndexBuffer: IDirect3DIndexBuffer9;

  hr: HRESULT;
  dwFaces, dwVertices: DWORD;

  pVertexBuffer: IDirect3DVertexBuffer9;

label
  FAIL;
begin
  pMeshSysMem:= nil;
  pMeshSysMem2:= nil;
  pIndexBuffer:= nil;

  pVertexBuffer:= nil;

  hr:= D3DXLoadMeshFromX(name, D3DXMESH_SYSTEMMEM,
                         m_pd3dDevice, nil, nil, nil, nil, pMeshSysMem);
  if FAILED(hr) then goto FAIL;

  //reform the mesh to one which will have space for the tangent vectors
  declVS[4]:= D3DDECL_END;
  hr := pMeshSysMem.CloneMesh(D3DXMESH_MANAGED, @declVS, m_pd3dDevice, pMeshSysMem2);
  if FAILED(hr) then goto FAIL;

  //compute the normals
  hr:= D3DXComputeNormals(pMeshSysMem, nil);
  if FAILED(hr) then goto FAIL;

  //compute the tangent vectors in the texture space and load them into
  //the tangent field (which is 3d), read the texture coords from the 0th texture
  //stage. Don't need the V direction vector since it is assumed to be U x N.
  hr := D3DXComputeTangent(pMeshSysMem2, 0, 0, 0, iTrue, nil);
  if FAILED(hr) then goto FAIL;

  SAFE_RELEASE(pMeshSysMem);

  hr := pMeshSysMem2.GetIndexBuffer(pIndexBuffer);
  if FAILED(hr) then goto FAIL;

  hr := pMeshSysMem2.GetVertexBuffer(pVertexBuffer);
  if FAILED(hr) then goto FAIL;

  dwFaces    := pMeshSysMem2.GetNumFaces;
  dwVertices := pMeshSysMem2.GetNumVertices;

  //give them to the shadow object
  m_pShadow.SetBuffers(pIndexBuffer, pVertexBuffer, dwFaces, dwVertices);

  SAFE_RELEASE(pVertexBuffer);
  SAFE_RELEASE(pIndexBuffer);

  //cleanup
FAIL:

  SAFE_RELEASE(pMeshSysMem);
  SAFE_RELEASE(pMeshSysMem2);
  SAFE_RELEASE(pVertexBuffer);
  SAFE_RELEASE(pIndexBuffer);
end;

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Animates the scene
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove: HResult;
var
//  matWorld, mat1, mat2: TD3DXMatrix;
  vecT, vecR, v1, v2: TD3DXVector3;

  matT, matR: TD3DXMatrix;
  qR: TD3DXQuaternion;
begin
  // Update world matrix

  // Process keyboard input
  vecT:= D3DXVector3(0.0, 0.0, 0.0);
  vecR:= D3DXVector3(0.0, 0.0, 0.0);

  if (m_bKey[Ord('W')]) then vecT.z:= vecT.z + 2.0; // Move Forward
  if (m_bKey[Ord('S')]) then vecT.z:= vecT.z - 2.0; // Move Backward

  // m_vecVelocity = m_vecVelocity * 0.9f - vecT * 0.1f;
  D3DXVec3Subtract(m_vecVelocity,
    D3DXVec3Scale(v1, m_vecVelocity, 0.9)^,
    D3DXVec3Scale(v2, vecT, 0.1)^);
  // m_vecAngularVelocity = m_vecAngularVelocity * 0.9f - vecR * 0.1f;
  D3DXVec3Subtract(m_vecAngularVelocity,
    D3DXVec3Scale(v1, m_vecAngularVelocity, 0.9)^,
    D3DXVec3Scale(v2, vecR, 0.1)^);

  // Update position and view matricies
  // m_vecVelocity = m_vecVelocity * 0.9f - vecT * 0.1f;
  D3DXVec3Scale(vecT, m_vecVelocity, m_fElapsedTime * m_fSpeed);
  // m_vecAngularVelocity = m_vecAngularVelocity * 0.9f - vecR * 0.1f;
  D3DXVec3Scale(vecR, m_vecAngularVelocity, m_fElapsedTime * m_fAngularSpeed);

  D3DXMatrixTranslation(matT, -vecT.x, -vecT.y, -vecT.z);

  D3DXMatrixMultiply(m_matPosition, matT, m_matPosition);

  D3DXQuaternionRotationYawPitchRoll(qR, vecR.y, vecR.x, vecR.z);
  D3DXMatrixRotationQuaternion(matR, qR);
  D3DXMatrixMultiply(m_matPosition, matR, m_matPosition);

  D3DXMatrixInverse(m_matView, nil, m_matPosition);

  //apply rotation
  if m_bRotate then
  begin
    D3DXMatrixRotationY(matR, -m_fElapsedTime*0.5);
     D3DXMatrixMultiply(m_matObject, matR, m_matObject);
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Renders the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Render: HResult;
var
  fAspectRatio: Single;
begin
  fAspectRatio := m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;

  // Update projection matrix
  if (m_fAspectRatio <> fAspectRatio) then
  begin
    D3DXMatrixPerspectiveFovLH(m_matProj, 0.05, fAspectRatio, 1.0, 10000.0);
    m_fAspectRatio := fAspectRatio;
  end;

  m_pShadow.ComputeLight(m_LightDir, m_matObject, m_matView, m_matProj);

  Result:= m_pd3dDevice.BeginScene;
  if FAILED(Result) then Exit;

  if m_bWindowed then
  begin
    m_pShadow.Render(m_bDebug, m_bShadow, m_bBump, m_rcWindowClient.left,
                     m_rcWindowClient.top, m_rcWindowClient.right - m_rcWindowClient.left,
                     m_rcWindowClient.bottom - m_rcWindowClient.top);
  end else
  begin
    m_pShadow.Render(m_bDebug, m_bShadow, m_bBump, 0, 0,
                     m_d3dsdBackBuffer.Width, m_d3dsdBackBuffer.Height);
  end;

  // Output statistics
  m_pFont.DrawText(2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats);
  m_pFont.DrawText(2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats);
  m_pFontSmall.DrawText(2, 40, D3DCOLOR_ARGB(255,255,255,255), '2: Toggle shadow');
  m_pFontSmall.DrawText(2, 60, D3DCOLOR_ARGB(255,255,255,255), '3: Toggle bumpmapping');
  m_pFontSmall.DrawText(2, 80, D3DCOLOR_ARGB(255,255,255,255), '4: Reset object');
  m_pFontSmall.DrawText(2,100, D3DCOLOR_ARGB(255,255,255,255), '5: Toggle spin');

  Result:= m_pd3dDevice.EndScene;
  if FAILED(Result) then Exit;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function CMyD3DApplication.OneTimeSceneInit: HResult;
var
  strBump: array[0..159] of Char;
  strColor: array[0..159] of Char;
begin
  // Set cursor to indicate that user can move the object with the mouse
{$IFDEF WIN64}
  SetClassLongPtr(m_hWnd, GCLP_HCURSOR, (LONG_PTR)LoadCursor(NULL, IDC_SIZEALL));
{$ELSE}
  SetClassLong(m_hWnd, GCL_HCURSOR, LoadCursor(0, IDC_SIZEALL));
{$ENDIF}

  Result := DXUtil_FindMediaFileCb(strBump, SizeOf(strBump), 'earthbump.bmp');
  if FAILED(Result) then Exit;
  Result := DXUtil_FindMediaFileCb(strColor, SizeOf(strColor), 'earth.bmp');
  if FAILED(Result) then Exit;

  m_pShadow := TShadowSet.Create(strBump, strColor);
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitDeviceObjects: HResult;
var
  matT: TD3DXMatrix;
  Caps: TD3DCaps9;
  strFile: array[0..159] of Char;
begin
  m_pFont.InitDeviceObjects(m_pd3dDevice);
  m_pFontSmall.InitDeviceObjects(m_pd3dDevice);

  D3DXMatrixIdentity(m_matObject);
  D3DXMatrixIdentity(m_matView);
  D3DXMatrixIdentity(m_matPosition);

  m_fSpeed             := 20.0;
  m_fAngularSpeed      := 0.10;
  m_vecVelocity        := D3DXVector3(0.0, 0.0, 0.0);
  m_vecAngularVelocity := D3DXVector3(0.0, 0.0, 0.0);
  m_LightDir           := D3DXVector3(0.0, 0.0, 1.0);

  D3DXMatrixTranslation(matT, 0,0, -40);

  D3DXMatrixMultiply(m_matPosition, matT, m_matPosition);
  D3DXMatrixInverse(m_matView, nil, m_matPosition);

  m_pd3dDevice.GetDeviceCaps(Caps);

  //detect if the hardware supports pixel shaders
  //if it does, then the shadow term can be done in one pass,
  //resulting in much faster and better results
  if (Caps.PixelShaderVersion < D3DPS_VERSION(1,1)) then
    m_bUsePix := False
  else
    m_bUsePix := True;

  Result := m_pShadow.Init(m_pd3dDevice, m_bUsePix);
  if FAILED(Result) then Exit;

  Result := DXUtil_FindMediaFileCb(strFile, SizeOf(strFile), 'sphere.x');
  if FAILED(Result) then Exit;
  LoadXFile(strFile);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Initialize scene objects
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects: HResult;
begin
  m_pFont.RestoreDeviceObjects;
  m_pFontSmall.RestoreDeviceObjects;

  m_pd3dDevice.SetRenderState(D3DRS_LIGHTING, iFalse);
  m_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iTrue);

  m_pShadow.Restore;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects: HResult;
begin
  m_pFont.InvalidateDeviceObjects;
  m_pFontSmall.InvalidateDeviceObjects;
  m_pShadow.Invalidate;
  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects: HResult;
begin
  m_pFont.DeleteDeviceObjects;
  m_pFontSmall.DeleteDeviceObjects;
  m_pShadow.DeleteDeviceObj;
  Result := S_OK;
end;

constructor TShadowSet.Create(HeightMapName, ColorTexName: PChar);
var
  i, x, y: DWord;
  fRange: Single;

  rgbq: array [0..255] of TRGBQuad;
  bp: BITMAP;
  bits: PByteArray;
  tb: PByteArray;
  texmap: HBITMAP;
  tDC: HDC;
  temploc, srcloc: PByteArray;

//  curHeight, curWidth: DWord;
label
  FAIL;
begin
  fRange := D3DX_PI / 4;

  m_bInitalize := False;
  m_bLoaded := False;

  //set everything to NULL
  for i:= 0 to 15 do
  begin
    m_pBasisMap[i] := nil;
    m_pHorizonMap[i] := nil;
    m_pHeightMap[i] := nil;
    if (i<4) then
    begin
      m_pBasisTextures[i] := nil;
      m_pHorizonTextures[i] := nil;
    end;
  end;

  m_pRenderBuffer  := nil;
  m_pRenderBuffer2 := nil;

  m_pNormalMap     := nil;
  m_pColorTexture  := nil;

  if Assigned(ColorTexName) then
    StrCopy(m_cColorName, ColorTexName)
  else
    m_cColorName[0] := #0;

  if Assigned(HeightMapName) then
    StrCopy(m_cHeightName, HeightMapName)
  else
    m_cColorName[0] := #0;

  //Load Bitmap
  texmap:= HBITMAP(LoadImage(0, HeightMapName, IMAGE_BITMAP, 0, 0,
                             LR_LOADFROMFILE or LR_CREATEDIBSECTION));
  if (texmap = 0) then goto FAIL;
  GetObject(texmap, SizeOf(BITMAP), @bp);

  m_iHeightWidth  := bp.bmWidth;
  m_iHeightHeight := bp.bmHeight;

  //XXX - need to  check for powers of 2
  tDC := CreateCompatibleDC(0);
  SelectObject(tDC, texmap);
  GetObject(tDC, SizeOf(BITMAP), @bp);
  bits := bp.bmBits;

  GetDIBColorTable(tDC, 0, 256, rgbq);

  GetMem(m_pHeightMap[0], m_iHeightWidth*m_iHeightHeight);
  tb := PByteArray(m_pHeightMap[0]);
  if(tb = nil) then goto FAIL;

  if (bp.bmBitsPixel = 8) then
  begin
    for y:= 0 to m_iHeightHeight - 1 do
    begin
      for x:= 0 to m_iHeightWidth - 1 do
      begin
        temploc := @tb[y * m_iHeightWidth + x];
        srcloc := @bits[(m_iHeightHeight-y-1) * DWord(bp.bmWidthBytes) + x];
        temploc[0] := rgbq[srcloc[0]].rgbRed;
      end;
    end;
  end else
  begin
    for y:= 0 to m_iHeightHeight - 1 do
    begin
      for x:= 0 to m_iHeightWidth - 1 do
      begin
        temploc := @tb[(y * (m_iHeightWidth) + x)];
        srcloc := @bits[(m_iHeightHeight-y-1) * DWord(bp.bmWidthBytes) + x*3];
        temploc[0] := srcloc[2];
      end;
    end;
  end;

  DeleteObject(texmap);
  DeleteDC(tDC);

{BAA: never used
  curHeight := m_iHeightHeight;
  curWidth  := m_iHeightWidth;

  curHeight := curHeight shr 1;
  curWidth  := curWidth  shr 1;
}
  //now set up the basis maps
  for i:= 0 to 7 do
  begin
    GetMem(m_pBasisMap[i], BASISMAPSIZE*BASISMAPSIZE);
    GetMem(m_pHorizonMap[i], m_iHeightWidth*m_iHeightHeight);

    if (m_pBasisMap[i] = nil) or (m_pHorizonMap[i] = nil) then goto FAIL;

    BuildBasisMap(m_pBasisMap[i], (D3DX_PI*i)/4.0, fRange);
    BuildHorizonMap(m_pHorizonMap[i], m_pHeightMap[0],
                    Cos((D3DX_PI*i)/4.0), Sin((D3DX_PI*i)/4.0), 15,
                    m_iHeightWidth, m_iHeightHeight, True, True);
  end;

  m_pShad1 := nil;
  m_pRenderBuffer := nil;
  m_pIndexBuffer := nil;
  m_pScreenVB := nil;

  m_iNumTriangles := 0;
  m_iNumVertices  := 0;
  m_pVertices := nil;

  m_pDev := nil;
  m_pTempBuffer := nil;
  m_pOldRenderTarget := nil;
  m_pVertexDeclaration := nil;
  m_pBasisShader := nil;
  m_pShadowShader := nil;
  m_pBumpShader := nil;
  m_pPixShader := nil;
  m_dwHeightLevels := 0;
  m_bUsePix := False;

  m_bLoaded := True;
  Exit;

FAIL:
  Exit;
end;

destructor TShadowSet.Destroy;
var
  i: Integer;
begin
  for i:= 0 to 15 do
  begin
    if Assigned(m_pBasisMap[i])   then FreeMem(m_pBasisMap[i]);
    if Assigned(m_pHorizonMap[i]) then FreeMem(m_pHorizonMap[i]);
    if Assigned(m_pHeightMap[i])  then FreeMem(m_pHeightMap[i]);
  end;
end;

var
  declVS2: array[0..4] of TD3DVertexElement9 =
  (
    (Stream: 0; Offset:  0; _Type: D3DDECLTYPE_FLOAT3; Method: D3DDECLMETHOD_DEFAULT; Usage: D3DDECLUSAGE_POSITION; UsageIndex: 0), // v0
    (Stream: 0; Offset: 12; _Type: D3DDECLTYPE_FLOAT3; Method: D3DDECLMETHOD_DEFAULT; Usage: D3DDECLUSAGE_NORMAL; UsageIndex: 0),   // v3
    (Stream: 0; Offset: 24; _Type: D3DDECLTYPE_FLOAT2; Method: D3DDECLMETHOD_DEFAULT; Usage: D3DDECLUSAGE_TEXCOORD; UsageIndex: 0), // v7
    (Stream: 0; Offset: 32; _Type: D3DDECLTYPE_FLOAT3; Method: D3DDECLMETHOD_DEFAULT; Usage: D3DDECLUSAGE_TANGENT; UsageIndex: 0),  // v8
    ({D3DDECL_END})
  );
function TShadowSet.Init(pDev: IDirect3DDevice9; bUsePix: BOOL): HResult;
var
  pCode: ID3DXBuffer;
  iWidth,iHeight: Integer;

  pHeightMap: IDirect3DTexture9;
  desc: TD3DSurfaceDesc;

  Shad1, Shad2, Shad3: array[0..254] of Char;
label
  FAIL;
begin
  pCode := nil;
  m_pDev := pDev;

  declVS2[4]:= D3DDECL_END;
  Result := m_pDev.CreateVertexDeclaration(@declVS2, m_pVertexDeclaration);
  if FAILED(Result) then Exit;

  m_bUsePix := bUsePix;
  if not m_bLoaded then goto FAIL;

  Result:= D3DXCreateTextureFromFile(m_pDev, m_cColorName, m_pColorTexture);
  if FAILED(Result) then goto FAIL;

  Result:= D3DXCreateTextureFromFile(m_pDev, m_cHeightName, pHeightMap);
  if FAILED(Result) then goto FAIL;

  pHeightMap.GetLevelDesc(0, desc);

  iWidth  := desc.Width;
  iHeight := desc.Height;

  Result:= D3DXCreateTexture(m_pDev, iWidth, iHeight, 0, 0, D3DFMT_A8R8G8B8,
                             D3DPOOL_MANAGED, m_pNormalMap);
  if FAILED(Result) then goto FAIL;

  D3DXComputeNormalMap(m_pNormalMap, pHeightMap, nil, 0, D3DX_CHANNEL_RED, 10);

  //two methods
  if not m_bUsePix then
  begin
    //don't have a dp4, so need to put the maps into 3 basis texture and 3 horizon maps
    //basis maps are not-content specific - they are just LERP tables

    //build the interleave maps
    if not BuildInterleavedMap(m_pBasisTextures[0], m_pBasisMap[0], m_pBasisMap[1], m_pBasisMap[2], nil, BASISMAPSIZE, BASISMAPSIZE)
      then goto FAIL;
    if not BuildInterleavedMap(m_pBasisTextures[1], m_pBasisMap[3], m_pBasisMap[4], m_pBasisMap[5], nil, BASISMAPSIZE, BASISMAPSIZE)
      then goto FAIL;
    if not BuildInterleavedMap(m_pBasisTextures[2], m_pBasisMap[6], m_pBasisMap[7], nil, nil, BASISMAPSIZE, BASISMAPSIZE)
      then goto FAIL;

    if not BuildInterleavedMap(m_pHorizonTextures[0], m_pHorizonMap[0], m_pHorizonMap[1], m_pHorizonMap[2], nil, m_iHeightWidth, m_iHeightHeight)
      then goto FAIL;
    if not BuildInterleavedMap(m_pHorizonTextures[1], m_pHorizonMap[3], m_pHorizonMap[4], m_pHorizonMap[5], nil, m_iHeightWidth, m_iHeightHeight)
      then goto FAIL;
    if not BuildInterleavedMap(m_pHorizonTextures[2], m_pHorizonMap[6], m_pHorizonMap[7], nil, nil, m_iHeightWidth, m_iHeightHeight)
      then goto FAIL;

    Result:= D3DXFilterTexture(m_pHorizonTextures[0], nil, 0, D3DX_FILTER_BOX);
    if FAILED(Result) then goto FAIL;
    Result:= D3DXFilterTexture(m_pHorizonTextures[1], nil, 0, D3DX_FILTER_BOX);
    if FAILED(Result) then goto FAIL;
    Result:= D3DXFilterTexture(m_pHorizonTextures[2], nil, 0, D3DX_FILTER_BOX);
    if FAILED(Result) then goto FAIL;

    Result := DXUtil_FindMediaFileCb(Shad1, SizeOf(Shad1), 'bumpshader.vsh');
    if FAILED(Result) then goto FAIL;
    Result := DXUtil_FindMediaFileCb(Shad2, SizeOf(Shad2), 'bumpshader2.vsh');
    if FAILED(Result) then goto FAIL;
    Result := DXUtil_FindMediaFileCb( Shad3, SizeOf(Shad3), 'bumpshader3.vsh');
    if FAILED(Result) then goto FAIL;

    Result:= D3DXAssembleShaderFromFile(Shad2, nil, nil, 0, @pCode, nil);
    if FAILED(Result) then goto FAIL;
    Result:= m_pDev.CreateVertexShader(pCode.GetBufferPointer, m_pBasisShader);
    if FAILED(Result) then goto FAIL;
    SAFE_RELEASE(pCode);

    Result:= D3DXAssembleShaderFromFile(Shad1, nil, nil, 0, @pCode, nil);
    if FAILED(Result) then goto FAIL;
    Result:= m_pDev.CreateVertexShader(pCode.GetBufferPointer, m_pShadowShader);
    if FAILED(Result) then goto FAIL;
    SAFE_RELEASE(pCode);

    Result:= D3DXAssembleShaderFromFile(Shad3, nil, nil, 0, @pCode, nil);
    if FAILED(Result) then goto FAIL;
    Result:= m_pDev.CreateVertexShader(pCode.GetBufferPointer, m_pBumpShader);
    if FAILED(Result) then goto FAIL;
    SAFE_RELEASE(pCode);

    //create render targets for pic-shader emulation
    Result:= m_pDev.CreateTexture(RENDERTARGSIZE, RENDERTARGSIZE, 0, D3DUSAGE_RENDERTARGET, D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, m_pRenderBuffer, nil);
    if FAILED(Result) then goto FAIL;

    Result:= m_pDev.CreateTexture(RENDERTARGSIZE, RENDERTARGSIZE, 0, D3DUSAGE_RENDERTARGET, D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, m_pRenderBuffer2, nil);
    if FAILED(Result) then goto FAIL;
  end else
  begin
    //build the interleave maps, use the Alpha to
    //basis maps are not-content specific - they are just LERP tables
    if not BuildInterleavedMap(m_pBasisTextures[0], m_pBasisMap[0], m_pBasisMap[1], m_pBasisMap[2], m_pBasisMap[3], BASISMAPSIZE, BASISMAPSIZE)
      then goto FAIL;
    if not BuildInterleavedMap(m_pBasisTextures[1], m_pBasisMap[4], m_pBasisMap[5], m_pBasisMap[6], m_pBasisMap[7], BASISMAPSIZE, BASISMAPSIZE)
      then goto FAIL;

    if not BuildInterleavedMap(m_pHorizonTextures[0], m_pHorizonMap[0], m_pHorizonMap[1], m_pHorizonMap[2], m_pHorizonMap[3], m_iHeightWidth, m_iHeightHeight)
      then goto FAIL;
    if not BuildInterleavedMap(m_pHorizonTextures[1], m_pHorizonMap[4], m_pHorizonMap[5], m_pHorizonMap[6],m_pHorizonMap[7], m_iHeightWidth, m_iHeightHeight)
      then goto FAIL;

    Result:= D3DXFilterTexture(m_pHorizonTextures[0], nil, 0, D3DX_FILTER_BOX);
    if FAILED(Result) then goto FAIL;
    Result:= D3DXFilterTexture(m_pHorizonTextures[1], nil, 0, D3DX_FILTER_BOX);
    if FAILED(Result) then goto FAIL;

    //load the required shaders
    Result := DXUtil_FindMediaFileCb(Shad1, SizeOf(Shad1), 'bumpshader4.vsh');
    if FAILED(Result) then goto FAIL;
    Result := DXUtil_FindMediaFileCb(Shad2, SizeOf(Shad2), 'shadowbumpshader.psh');
    if FAILED(Result) then goto FAIL;
    Result := DXUtil_FindMediaFileCb( Shad3, SizeOf(Shad3), 'bumpshader3.vsh');
    if FAILED(Result) then goto FAIL;

    Result:= D3DXAssembleShaderFromFile(Shad1, nil, nil, 0, @pCode, nil);
    if FAILED(Result) then goto FAIL;
    Result:= m_pDev.CreateVertexShader(pCode.GetBufferPointer, m_pShadowShader);
    if FAILED(Result) then goto FAIL;
    SAFE_RELEASE(pCode);

    Result:= D3DXAssembleShaderFromFile(Shad2, nil, nil, 0, @pCode, nil);
    if FAILED(Result) then goto FAIL;
    Result:= m_pDev.CreatePixelShader(pCode.GetBufferPointer, m_pPixShader);
    if FAILED(Result) then goto FAIL;
    SAFE_RELEASE(pCode);

    Result:= D3DXAssembleShaderFromFile(Shad3, nil, nil, 0, @pCode, nil);
    if FAILED(Result) then goto FAIL;
    Result:= m_pDev.CreateVertexShader(pCode.GetBufferPointer, m_pBumpShader);
    if FAILED(Result) then goto FAIL;
    SAFE_RELEASE(pCode);
  end;

  m_bInitalize := True;

  Result:= S_OK;
  Exit;
FAIL:

  SAFE_RELEASE(pCode);
  Result:= E_FAIL;
end;

procedure TShadowSet.DeleteDeviceObj;
var
  i: Integer;
begin
  for i:= 0 to 3 do
  begin
    SAFE_RELEASE(m_pHorizonTextures[i]);
    SAFE_RELEASE(m_pBasisTextures[i]);
  end;
  SAFE_RELEASE(m_pNormalMap);
  SAFE_RELEASE(m_pColorTexture);
  SAFE_RELEASE(m_pRenderBuffer);
  SAFE_RELEASE(m_pRenderBuffer2);
  SAFE_RELEASE(m_pIndexBuffer);
  SAFE_RELEASE(m_pVertices);
  SAFE_RELEASE(m_pScreenVB);
  SAFE_RELEASE(m_pShad1);
  SAFE_RELEASE(m_pVertexDeclaration);
  SAFE_RELEASE(m_pBumpShader);
  SAFE_RELEASE(m_pShadowShader);
  SAFE_RELEASE(m_pBasisShader);
  SAFE_RELEASE(m_pPixShader);
  
  m_pDev := nil; // Delphi version
end;

procedure TShadowSet.Restore;
begin
  SAFE_RELEASE(m_pRenderBuffer);
  SAFE_RELEASE(m_pRenderBuffer2);

  if not m_bUsePix then
  begin
    m_pDev.CreateTexture(RENDERTARGSIZE, RENDERTARGSIZE, 0,
      D3DUSAGE_RENDERTARGET, D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, m_pRenderBuffer, nil);
    m_pDev.CreateTexture(RENDERTARGSIZE, RENDERTARGSIZE, 0,
      D3DUSAGE_RENDERTARGET, D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, m_pRenderBuffer2, nil);
  end;

  m_pDev.GetRenderTarget(0, m_pOldRenderTarget);
end;

procedure TShadowSet.Invalidate;
begin
  SAFE_RELEASE(m_pRenderBuffer);
  SAFE_RELEASE(m_pRenderBuffer2);
  SAFE_RELEASE(m_pOldRenderTarget);
end;

// this builds a basis map
// the basis maps will be interleaved, or we will pack 3 of them into one texture
// (we could pack 4, but we can't use the alpha channel in our dot multiplies)
// but we will assume that the u,v represent the angle with respect to the center of the
// of the texture (.5,.5), So if we want the basis function for the light vector
// (1,.5) we'd look at the point (1,.75)
procedure TShadowSet.BuildBasisMap(pBasis: PByteArray; PrimAngle, fAngle2: Single);
var
  x, y: Integer;
  fEndX, fEndY, fAngle, fPercent: Single;
  u, v, dot, nx, ny, fsq: Single;
  pCurLine: PByte;
begin
  u := Cos(PrimAngle);
  v := Sin(PrimAngle);

  for y:= 0 to BASISMAPSIZE - 1 do
  begin
    pCurLine := @pBasis[y*BASISMAPSIZE];

    for x:= 0 to BASISMAPSIZE - 1 do
    begin
      fEndX := x - 0.5*BASISMAPSIZE;
      fEndY := y - 0.5*BASISMAPSIZE;

      //take the dot product of the normalized vectors
      fsq := Sqrt(fEndX*fEndX + fEndY*fEndY);
      if (fsq = 0) then
      begin
        fPercent := 128;
      end else
      begin
        // we remember our definiton of dot product,
        // cos(Angle) = DotProduct of Normalzed vectors
        // so Angle = acos(DotProduct)
        nx := fEndX/fsq;
        ny := fEndY/fsq;
        dot := nx*u + ny*v;
        if (dot < -1.0) then dot := -1.0;
        if (dot >  1.0) then dot := 1.0;
        fAngle := ArcCos(dot);
        if (Abs(fAngle) < fAngle2) then 
        begin
          fPercent := 128 + Abs(fAngle-fAngle2)*127/fAngle2;
          if (fPercent > 255) then fPercent:= 255;
        end else
        begin
          fPercent := 128;
        end;
      end;
      pCurLine^:= Round(fPercent);
      Inc(pCurLine);
    end;
  end;
end;

function TShadowSet.BuildInterleavedMap(out newTex: IDirect3DTexture9;
  r, g, b, a: PByteArray; width, height: Integer): Boolean;
var
  hr: HResult;
  pTexture: IDirect3DTexture9;
  LockedRect: TD3DLockedRect;
  pBase, pLine: PByteArray;
  x, y: Integer;
begin
  hr:= D3DXCreateTexture(m_pDev, width, height, 0, 0, D3DFMT_A8R8G8B8, D3DPOOL_MANAGED, pTexture);
  if (hr <> S_OK) then
  begin
    Result:= False;
    Exit;
  end;

  hr := pTexture.LockRect(0, LockedRect, nil, 0);
  if (hr <> S_OK) then
  begin
    Result:= False;
    Exit;
  end;

  pBase := LockedRect.pBits;

  for y:= 0 to height - 1 do
  begin
    pLine := @pBase[y*LockedRect.Pitch];
    for x:= 0 to width - 1 do
    begin
      if Assigned(r) then
        pLine[2] := r[y*width + x]
      else
        pLine[2] := 128;

      if Assigned(g) then
        pLine[1] := g[y*width + x]
      else
        pLine[1] := 128;

      if Assigned(b) then
        pLine[0] := b[y*width + x]
      else
        pLine[0] := 128;

      if Assigned(a) then
        pLine[3] := a[y*width + x]
      else
        pLine[3] := 128;

      pLine:= Pointer(Integer(pLine) + 4);
    end;
  end;

  pTexture.UnlockRect(0);

  newTex := pTexture;

  Result:= TRUE;
end;

const
  MAPHEIGHT  = 0.175;
  MAPHEIGHT2 = 0.175;

procedure TShadowSet.BuildHorizonMap(pHor, pHeight: PByteArray; dx, dy: Single;
  iStepLength: Integer; iWidth, iHeight: Integer; bWrapU, bWrapV: Boolean);
var
  x, y: Integer;
  imx, imy, i: Integer;
  fpx, fpy: Single;
  pCurHurLine: PByte;
  fMaxAngle, fNewAngle, fDeltaX, fDeltaY, fHeight: Single;
begin

  for y:= 0 to iHeight - 1 do
  begin
    pCurHurLine := @pHor[y*iWidth];
    for x:= 0 to iWidth - 1 do
    begin
      //we assume its completly visible at zero
      fMaxAngle := 0;
      fpx := x;
      fpy := y;
      for i:= 0 to iStepLength - 1 do
      begin
        fpx:= fpx + dx;
        fpy:= fpy + dy;

        imx := Round(fpx);
        imy := Round(fpy);
        //don't check against ourselves for blocking
        if (imx = x) and (imy = y) then
          Continue;
        //if we are out of bounds, we're done
        //if the texture is repeating - this should
        //probally wrap
        if (imx < 0) then 
        begin
          if bWrapU then
            imx := iWidth - 1
          else Continue;
        end;

        if (imx >= iWidth) then
        begin
          if bWrapU then
           imx := 0
          else Continue;
        end;

        if (imy < 0) then
        begin
          if bWrapV then
           imy := iHeight - 1
          else Continue;
        end;

        if (imy >= iHeight) then
        begin
          if bWrapV then
            imy := 0
          else Continue;
        end;

        fDeltaX := fpx - x;
        fDeltaY := fpy - y;
        fHeight := pHeight[imy*iWidth+imx]*MAPHEIGHT - pHeight[y*iWidth+x]*MAPHEIGHT;

        //this is the dot prodoct
        //Since we are dotting against the normal, (0,0,1) this is just
        // the Height normalized, or the Height over the length of the vector
        fNewAngle := fHeight/Sqrt(fDeltaX*fDeltaX + fDeltaY*fDeltaY + fHeight*fHeight);

        //if we found a further obstruction, use it
        if (fNewAngle > fMaxAngle) then
          fMaxAngle:= fNewAngle;
      end;

      //centered around 128
      pCurHurLine^:= 128 + Round(fMaxAngle*127);
      Inc(pCurHurLine);
    end;
  end;
end;

procedure TShadowSet.ComputeLight(const lightDir: TD3DXVector3;
  const pObject, pView, pProj: TD3DXMatrix);
var
//  U, V, W, Pos: TD3DXVector3;
  light, one, half: TD3DXVector4;
  temp: TD3DXMatrix;
begin
  one:= D3DXVector4(1, 1, 1, 1);
  half:= D3DXVector4(0.5, 0.5, 0.5, 0.5);

  if not m_bInitalize then Exit;

  light.x := lightDir.x;
  light.y := lightDir.y;
  light.z := lightDir.z;
  light.w := 0;

  D3DXVec4Normalize(light, light);

  m_matObject := pObject;
  m_matView   := pView;
  m_matProj   := pProj;

  D3DXMatrixMultiply(m_matTot, pObject, pView);

  D3DXMatrixTranspose(temp, m_matTot);

  //the inverse transform
  m_pDev.SetVertexShaderConstantF(16, @pObject, 4);

  D3DXMatrixMultiply(m_matTot, m_matTot, pProj);

  D3DXMatrixTranspose(m_matTot, m_matTot);
  D3DXMatrixTranspose(m_matObject, pObject);

  //load the values into the registers
  m_pDev.SetVertexShaderConstantF(12, @light, 1);
  m_pDev.SetVertexShaderConstantF(32, @one,   1);
  m_pDev.SetVertexShaderConstantF(33, @half,  1);
  m_pDev.SetVertexShaderConstantF( 0, @m_matObject, 4);
  m_pDev.SetVertexShaderConstantF( 8, @m_matTot, 4);
end;

procedure TShadowSet.Render(bShow, bShad, bBump: BOOL; iLeft, iTop, iRight, iBottom: Integer);
var
  i: Integer;
  pVerts: PByte;
  pSurface: IDirect3DSurface9;
  ViewPort, OldPort: TD3DViewport9;

type
  T1 = packed record
    x,y,z,w,
    u1,v1: Single;
  end;
  T2 = packed record
    x,y,z,w: Single;
    color: DWord;
  end;
{$J+} // {$WRITEABLECONST ON}
const
  s_Verts2: array[0..3] of T1 =
  (
    (x: RENDERTARGSIZE; y: 0.0;            z: 0.0; w: 1.0;   u1:1; v1:0),
    (x: RENDERTARGSIZE; y: RENDERTARGSIZE; z: 0.0; w: 1.0;   u1:1; v1:1),
    (x: 0.0;            y: 0.0;            z: 0.0; w: 1.0;   u1:0; v1:0),
    (x: 0.0;            y: RENDERTARGSIZE; z: 0.0; w: 1.0;   u1:0; v1:1)
  );

  s_Verts3: array[0..3] of T2 =
  (
    (x: 750.0; y:   0.0; z: 0.0; w: 0{, D3DCOLOR_ARGB( 0xa0,0xff, 0x00, 0x00 )}),
    (x: 750.0; y: 375.0; z: 0.0; w: 0{, D3DCOLOR_ARGB( 0xa0,0xff,00, 0x00)}),
    (x:   0.0; y:   0.0; z: 0.0; w: 0{, D3DCOLOR_ARGB( 0xa0,0xff,0x00, 0x00)}),
    (x:   0.0; y: 375.0; z: 0.0; w: 0{, D3DCOLOR_ARGB( 0xa0,0xff, 0x00, 0x00)})
  );
{$J-} // {$WRITEABLECONST OFF}
label
  FAIL;
begin
  if not Assigned(m_pVertices) or not Assigned(m_pIndexBuffer) or not Assigned(m_pDev)
  then Exit;

  if not m_bInitalize then Exit;

  s_Verts3[0].color:= D3DCOLOR_ARGB($a0, $ff, 0, 0);
  s_Verts3[1].color:= D3DCOLOR_ARGB($a0, $ff, 0, 0);
  s_Verts3[2].color:= D3DCOLOR_ARGB($a0, $ff, 0, 0);
  s_Verts3[3].color:= D3DCOLOR_ARGB($a0, $ff, 0, 0);

  m_pDev.SetIndices(m_pIndexBuffer);

  if not m_bUsePix then
  begin
    //sets up polygons that
    //are the same size as the screen
    //(well... almost)
    OldPort.X := iLeft;
    OldPort.Y := iTop;
    OldPort.Width := iRight-iLeft;
    OldPort.Height := iBottom-iTop;
    OldPort.MaxZ := 1;
    OldPort.MinZ := 0;

    s_Verts2[0].x := OldPort.Width;
    s_Verts2[1].x := OldPort.Width;
    s_Verts2[1].y := OldPort.Height;
    s_Verts2[3].y := OldPort.Height;

    s_Verts3[0].x := OldPort.Width;
    s_Verts3[1].x := OldPort.Width;
    s_Verts3[1].y := OldPort.Height;
    s_Verts3[3].y := OldPort.Height;

    if not bShow then
    begin
      s_Verts3[0].color := D3DCOLOR_ARGB($a0, 0, 0, 0);
      s_Verts3[1].color := D3DCOLOR_ARGB($a0, 0, 0, 0);
      s_Verts3[2].color := D3DCOLOR_ARGB($a0, 0, 0, 0);
      s_Verts3[3].color := D3DCOLOR_ARGB($a0, 0, 0, 0);
    end else
    begin
      s_Verts3[0].color := D3DCOLOR_ARGB($70, $ff, 0, 0);
      s_Verts3[1].color := D3DCOLOR_ARGB($70, $ff, 0, 0);
      s_Verts3[2].color := D3DCOLOR_ARGB($70, $ff, 0, 0);
      s_Verts3[3].color := D3DCOLOR_ARGB($70, $ff, 0, 0);
    end;

    //fill the VBs
    if (m_pShad1 = nil) then 
    begin
      if FAILED(m_pDev.CreateVertexBuffer(SizeOf(s_Verts2), 0,
                         D3DFVF_XYZRHW or D3DFVF_TEX1, D3DPOOL_SYSTEMMEM, m_pShad1, nil))
      then goto Fail;
    end;

    if FAILED(m_pShad1.Lock(0, SizeOf(s_Verts2), Pointer(pVerts), 0)) then
      goto Fail;

    Move(s_Verts2, pVerts^, SizeOf(s_Verts2));

    if FAILED(m_pShad1.Unlock) then
      goto Fail;

    if (m_pScreenVB = nil) then
    begin
      if FAILED(m_pDev.CreateVertexBuffer(SizeOf(s_Verts3), 0,
                         D3DFVF_XYZRHW or D3DFVF_DIFFUSE, D3DPOOL_SYSTEMMEM, m_pScreenVB, nil))
      then goto Fail;
    end;

    if FAILED(m_pScreenVB.Lock(0, SizeOf(s_Verts3), Pointer(pVerts), 0)) then
      goto Fail;

    Move(s_Verts3, pVerts^, SizeOf(s_Verts3));

    if FAILED(m_pScreenVB.Unlock) then

    m_pDev.SetRenderState(D3DRS_ZENABLE, D3DZB_FALSE);

    if bShad then
    begin
     //first set the Rendertarget
      m_pRenderBuffer.GetSurfaceLevel(0, pSurface);
      m_pDev.SetRenderTarget(0, pSurface);

      ViewPort.X := 0;
      ViewPort.Y := 0;
      ViewPort.Height := RENDERTARGSIZE;
      ViewPort.Width  := RENDERTARGSIZE;
      ViewPort.MaxZ := 1;
      ViewPort.MinZ := 0;

      m_pDev.SetViewport(ViewPort);

      m_pDev.Clear(0, nil, D3DCLEAR_TARGET , D3DCOLOR_ARGB(0, 0, 0, 0), 1.0, 0);

      m_pDev.SetRenderState(D3DRS_ALPHABLENDENABLE, iTrue);
      m_pDev.SetRenderState(D3DRS_ALPHATESTENABLE,  iFalse);

      m_pDev.SetRenderState(D3DRS_WRAP0,D3DWRAPCOORD_0 or D3DWRAPCOORD_1);
      m_pDev.SetRenderState(D3DRS_WRAP1,D3DWRAPCOORD_0 or D3DWRAPCOORD_1);

      m_pDev.SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
      m_pDev.SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
      m_pDev.SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR);

      m_pDev.SetSamplerState(1, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
      m_pDev.SetSamplerState(1, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
      m_pDev.SetSamplerState(1, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR);

      m_pDev.SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, 0);
      m_pDev.SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_SELECTARG1);
      m_pDev.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
      m_pDev.SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG1);
      m_pDev.SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);

      m_pDev.SetTextureStageState(1, D3DTSS_TEXCOORDINDEX, 1);
      m_pDev.SetTextureStageState(1, D3DTSS_COLOROP, D3DTOP_DOTPRODUCT3);
      m_pDev.SetTextureStageState(1, D3DTSS_COLORARG1, D3DTA_TEXTURE);
      m_pDev.SetTextureStageState(1, D3DTSS_COLORARG2, D3DTA_CURRENT);
      m_pDev.SetTextureStageState(1, D3DTSS_ALPHAOP, D3DTOP_DOTPRODUCT3);
      m_pDev.SetTextureStageState(1, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);
      m_pDev.SetTextureStageState(1, D3DTSS_ALPHAARG2, D3DTA_CURRENT);

      m_pDev.SetTextureStageState(2, D3DTSS_COLOROP, D3DTOP_DISABLE);
      m_pDev.SetTextureStageState(2, D3DTSS_ALPHAOP, D3DTOP_DISABLE);

      m_pDev.SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);
      m_pDev.SetRenderState(D3DRS_SRCBLEND, D3DBLEND_ONE);

      m_pDev.SetRenderState(D3DRS_WRAP0, D3DWRAPCOORD_0 or D3DWRAPCOORD_1);
      m_pDev.SetRenderState(D3DRS_WRAP1, D3DWRAPCOORD_0 or D3DWRAPCOORD_1);

      //render all the basis functions times all the images
      m_pDev.SetStreamSource(0, m_pVertices, 0, SizeOf(TShaderVertex));
      m_pDev.SetVertexShader(m_pBasisShader);
      m_pDev.SetVertexDeclaration(m_pVertexDeclaration);
      for i:= 0 to 2  do
      begin
        m_pDev.SetTexture(0, m_pHorizonTextures[i]);
        m_pDev.SetTexture(1, m_pBasisTextures[i]);
        m_pDev.DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 0, m_iNumVertices, 0, m_iNumTriangles);
      end;

      m_pDev.SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_NONE);


      //now we have to render a simple alpha texture to subtract off for the shadow
      m_pRenderBuffer2.GetSurfaceLevel(0, pSurface);
      m_pDev.SetRenderTarget(0, pSurface);

      m_pDev.Clear(0, nil, D3DCLEAR_TARGET, D3DCOLOR_ARGB(0, 0, 0, 0), 1.0, 0);

      m_pDev.SetStreamSource(0, m_pVertices, 0, SizeOf(TShaderVertex));
      m_pDev.SetVertexShader(m_pShadowShader);
      m_pDev.SetVertexDeclaration(m_pVertexDeclaration);

      m_pDev.SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, 0);
      m_pDev.SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_SELECTARG1);
      m_pDev.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_DIFFUSE);
      m_pDev.SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG1);
      m_pDev.SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_DIFFUSE);

      m_pDev.SetTextureStageState(1, D3DTSS_COLOROP, D3DTOP_DISABLE);
      m_pDev.SetTextureStageState(1, D3DTSS_ALPHAOP, D3DTOP_DISABLE);

      m_pDev.SetRenderState(D3DRS_ALPHABLENDENABLE, iFalse);
      m_pDev.SetRenderState(D3DRS_ALPHATESTENABLE, iFalse);

      m_pDev.DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 0, m_iNumVertices, 0, m_iNumTriangles);

      //restore our states
      m_pDev.SetRenderTarget(0, m_pOldRenderTarget);
      pSurface:= nil;
      m_pDev.SetViewport(OldPort);
    end;
  end;

  m_pDev.Clear(0, nil, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER, D3DCOLOR_ARGB(0, 0, 0, 0), 1.0, 0);
  m_pDev.SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
  m_pDev.SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
  m_pDev.SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR);

  m_pDev.SetSamplerState(1, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
  m_pDev.SetSamplerState(1, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
  m_pDev.SetSamplerState(1, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR);

  m_pDev.SetRenderState(D3DRS_WRAP0,D3DWRAPCOORD_0 or D3DWRAPCOORD_1);
  m_pDev.SetRenderState(D3DRS_WRAP1,D3DWRAPCOORD_0 or D3DWRAPCOORD_1);
  //important to unset a pixel shader!
  m_pDev.SetPixelShader(nil);

  //diffuse, DOT3 lighting.
  if bBump then
  begin
    m_pDev.SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR);
    m_pDev.SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, 0);
    m_pDev.SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_DOTPRODUCT3);
    m_pDev.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
    m_pDev.SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
    m_pDev.SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG1);
    m_pDev.SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);

    m_pDev.SetSamplerState(1, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR);
    m_pDev.SetTextureStageState(1, D3DTSS_TEXCOORDINDEX, 1);
    m_pDev.SetTextureStageState(1, D3DTSS_COLOROP, D3DTOP_MODULATE);
    m_pDev.SetTextureStageState(1, D3DTSS_COLORARG1, D3DTA_TEXTURE);
    m_pDev.SetTextureStageState(1, D3DTSS_COLORARG2, D3DTA_CURRENT);
    m_pDev.SetTextureStageState(1, D3DTSS_ALPHAOP, D3DTOP_SELECTARG1);
    m_pDev.SetTextureStageState(1, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);

    m_pDev.SetTexture(1, m_pColorTexture);
    m_pDev.SetTexture(0, m_pNormalMap);
  end else
  begin
    //no lighting, just base color texture
    m_pDev.SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR);
    m_pDev.SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, 0);
    m_pDev.SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_SELECTARG1);
    m_pDev.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
    m_pDev.SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
    m_pDev.SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG1);
    m_pDev.SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);

    m_pDev.SetTextureStageState(1, D3DTSS_COLOROP, D3DTOP_DISABLE);
    m_pDev.SetTextureStageState(1, D3DTSS_ALPHAOP, D3DTOP_DISABLE);

    m_pDev.SetTexture(0, m_pColorTexture);
  end;

  m_pDev.SetStreamSource(0, m_pVertices, 0, SizeOf(TShaderVertex));

  m_pDev.SetRenderState(D3DRS_ALPHABLENDENABLE, iFalse);
  m_pDev.SetVertexShader(m_pBumpShader);
  m_pDev.SetVertexDeclaration(m_pVertexDeclaration);
  m_pDev.DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 0, m_iNumVertices, 0, m_iNumTriangles);

  m_pDev.SetTextureStageState(2, D3DTSS_COLOROP, D3DTOP_DISABLE);
  m_pDev.SetTextureStageState(2, D3DTSS_ALPHAOP, D3DTOP_DISABLE);
  m_pDev.SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_NONE);
  m_pDev.SetSamplerState(1, D3DSAMP_MIPFILTER, D3DTEXF_NONE);

  //now apply the black mask
  if bShad then
  begin
    //emmulate pixel shaders
    //the first two passes (the basis computation and the
    //light angle computation) were done before
    if not m_bUsePix then
    begin
      //now, subtract the light angle from the she shadow angle
      //if the value is non zero, then the pixel is in shaddow
      m_pDev.SetStreamSource(0, m_pShad1, 0, 6*SizeOf(Single));
      m_pDev.SetVertexShader(nil);
      m_pDev.SetFVF(D3DFVF_XYZRHW or D3DFVF_TEX1);

      m_pDev.SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, 0);

      m_pDev.SetRenderState(D3DRS_WRAP0, 0);
      m_pDev.SetRenderState(D3DRS_WRAP1, 0);

      m_pDev.SetRenderState(D3DRS_TEXTUREFACTOR, s_Verts3[0].color);
      m_pDev.SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_SELECTARG1);
      m_pDev.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
      m_pDev.SetTextureStageState(0, D3DTSS_ALPHAOP,D3DTOP_SELECTARG1);
      m_pDev.SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);

      if bShow then
        m_pDev.SetRenderState(D3DRS_TEXTUREFACTOR, D3DCOLOR_ARGB($ff,$ff, 0, 0))
      else
        m_pDev.SetRenderState(D3DRS_TEXTUREFACTOR, $80202020);

      m_pDev.SetRenderState(D3DRS_TEXTUREFACTOR, $60606060);
      m_pDev.SetTextureStageState(1, D3DTSS_COLOROP, D3DTOP_SELECTARG1);
      m_pDev.SetTextureStageState(1, D3DTSS_COLORARG1, D3DTA_TFACTOR);
      m_pDev.SetTextureStageState(1, D3DTSS_ALPHAOP, D3DTOP_SUBTRACT);
      m_pDev.SetTextureStageState(1, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);
      m_pDev.SetTextureStageState(1, D3DTSS_ALPHAARG2, D3DTA_CURRENT);

      //use an alpha test - if alpha is non zero,
      m_pDev.SetRenderState(D3DRS_ALPHATESTENABLE, iTrue);
      m_pDev.SetRenderState(D3DRS_ALPHABLENDENABLE, iTrue);
      m_pDev.SetRenderState(D3DRS_ALPHAFUNC, D3DCMP_NOTEQUAL);
      m_pDev.SetRenderState(D3DRS_ALPHAREF, 0);

      m_pDev.SetTextureStageState(1, D3DTSS_TEXCOORDINDEX, 0);
      m_pDev.SetRenderState(D3DRS_DESTBLEND,D3DBLEND_ZERO);
      m_pDev.SetRenderState(D3DRS_SRCBLEND,D3DBLEND_DESTCOLOR);

      m_pDev.SetTexture(0, m_pRenderBuffer2);
      m_pDev.SetTexture(1, m_pRenderBuffer);
      m_pDev.DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);

      m_pDev.SetTextureStageState(1, D3DTSS_TEXCOORDINDEX, 1);
    end else
    begin
       //use the pixel shader, much faster...

       //turn wrapping and blending on all the stages
       m_pDev.SetRenderState(D3DRS_WRAP0,D3DWRAPCOORD_0 or D3DWRAPCOORD_1);
       m_pDev.SetRenderState(D3DRS_WRAP1,D3DWRAPCOORD_0 or D3DWRAPCOORD_1);

       m_pDev.SetRenderState(D3DRS_WRAP2,D3DWRAPCOORD_0 or D3DWRAPCOORD_1);
       m_pDev.SetRenderState(D3DRS_WRAP3,D3DWRAPCOORD_0 or D3DWRAPCOORD_1);

       m_pDev.SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
       m_pDev.SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
       m_pDev.SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR);

       m_pDev.SetSamplerState(1, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
       m_pDev.SetSamplerState(1, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
       m_pDev.SetSamplerState(1, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR);

       m_pDev.SetSamplerState(2, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
       m_pDev.SetSamplerState(2, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
       m_pDev.SetSamplerState(2, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR);

       m_pDev.SetSamplerState(3, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
       m_pDev.SetSamplerState(3, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
       m_pDev.SetSamplerState(3, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR);

       //bind the textures
       m_pDev.SetTexture(0, m_pHorizonTextures[0]);
       m_pDev.SetTexture(1, m_pBasisTextures[0]);
       m_pDev.SetTexture(2, m_pHorizonTextures[1]);
       m_pDev.SetTexture(3, m_pBasisTextures[1]);

       //render all the basis functions times all the images
       m_pDev.SetStreamSource(0, m_pVertices, 0, SizeOf(TShaderVertex));
       m_pDev.SetVertexDeclaration(m_pVertexDeclaration);
       m_pDev.SetVertexShader(m_pShadowShader);
       m_pDev.SetPixelShader(m_pPixShader);

       m_pDev.SetRenderState(D3DRS_ALPHATESTENABLE, iFalse);
       m_pDev.SetRenderState(D3DRS_ALPHABLENDENABLE, iTrue);
       m_pDev.SetRenderState(D3DRS_DESTBLEND,D3DBLEND_INVSRCCOLOR);
       m_pDev.SetRenderState(D3DRS_SRCBLEND,D3DBLEND_ZERO);

       //run the pixel shader/vertex shader
       //these will summ up all the basis maps and then compare at a per
       //pixel level with the light angle.
       //the output is the (inverse) shadow term
       m_pDev.DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 0, m_iNumVertices, 0, m_iNumTriangles);
    end;
  end;
Fail:
end;




//set the index and vertex buffer for the object being rendered.
procedure TShadowSet.SetBuffers(pIndex: IDirect3DIndexBuffer9;
  pVertBuffer: IDirect3DVertexBuffer9; iNumFaces, iNumVertices: Integer);
begin
  m_pIndexBuffer := pIndex;

  m_pVertices    := pVertBuffer;

  m_iNumTriangles := iNumFaces;
  m_iNumVertices  := iNumVertices;
end;

end.

